

LOAD_DOCKER_CONTAINER = """
#!/usr/bin/env bash

set -e

if [[ "$(docker images -q {{image}} 2> /dev/null)" == "" ]]; then
    exec docker load -i /tmp/{{image_filename}}
fi
"""

START_DOCKER_CONTAINER = """
#!/usr/bin/env bash

set -e

INSTANCE_ID=$(curl -s http://169.254.169.254/latest/meta-data/instance-id)
REGION=$(curl -s http://169.254.169.254/latest/dynamic/instance-identity/document | grep region | awk -F\\" '{print $4}')

STACK_ENV=`aws ec2 describe-tags --filters "Name=resource-id,Values=$INSTANCE_ID" "Name=key,Values=Environment" --region=$REGION --output=text | awk '{print $5}'`

IFS=$'\r\n' GLOBIGNORE='*' command eval  'docker_args=($(cat /opt/nova/environments/$STACK_ENV/docker-vars.list))'

eval docker run -d \
  --name={{service_name}} \
  -p {{port}}:{{port}} \
  $(< /opt/nova/environments/$STACK_ENV/docker-opts.list) \
  ${docker_args[@]} \
  $(< /opt/nova/environments/$STACK_ENV/docker-vols.list) \
  {{image}} \
  $(< /opt/nova/environments/$STACK_ENV/docker-args.list)
"""

VALIDATE = """
#!/bin/bash

set -x

n=0
until [ $n -ge 30 ]
do
    sleep 10;
    res=`curl -s -I localhost:{{port}}{{healthcheck_url}} | grep HTTP/1.1 | awk {'print $2'}`
    echo $res;

    if [ $res -eq 200 ]
     then
      exit 0;
    fi
    n=$[$n+1]
done

exit 1;
"""



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from nova.core.managers.aws_manager import AwsManager
from nova.core.managers.docker_manager import DockerManager
from nova.core.managers.version_manager import VersionManager


class ManagerProvider(object):

    def __init__(self):
        pass

    @staticmethod
    def aws_manager(profile, region):
        return AwsManager(profile, region)

    @staticmethod
    def docker_manager():
        return DockerManager()

    @staticmethod
    def version_manager():
        return VersionManager()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
from collections import OrderedDict

from nova.core.exc import NovaError
from nova.core.spec import resource_name
from nova.core.spec.stack import Stack
from nova.core.cfn_pyplates.core import Resource, JSONableDict, Output
from nova.core.cfn_pyplates.functions import ref


class Environment(object):

    def __init__(self, name, aws_profile, aws_region, resources, deploy_arn, deployment_bucket, deployment_application_id, stacks):
        self.name = name
        self.aws_profile = aws_profile
        self.aws_region = aws_region
        self.resources = resources
        self.deploy_arn = deploy_arn
        self.deployment_bucket = deployment_bucket
        self.deployment_application_id = deployment_application_id
        self.stacks = stacks

    def yaml(self):
        data = OrderedDict([
            ('name', self.name),
            ('aws_profile', self.aws_profile),
            ('aws_region', self.aws_region),
            ('resources', self.resources),
            ('deploy_arn', self.deploy_arn),
            ('deployment_bucket', self.deployment_bucket),
            ('deployment_application_id', self.deployment_application_id),
            ('stacks', [s.yaml() for s in self.stacks])
        ])
        return OrderedDict((k, v) for k, v in data.items() if v is not None)

    def get_stack(self, stack_name):
        stack = next((item for item in self.stacks if item.name == stack_name), None)
        if stack is not None:
            return stack
        else:
            raise NovaError("Stack '%s' was not found in the configuration" % stack_name)

    def get_stack_templates_used(self, service, template_bucket):
        templates_used = dict()
        for s in self.stacks:
            templates_used[s.name] = s.get_template_used(service, template_bucket)
        return templates_used

    def to_cfn_template(self, service, template_bucket, cft, aws_manager, include_docker=True):
        extra_resources = self.__read_extra_resources()
        if extra_resources is not None:
            for r in extra_resources:
                cft.resources.add(r)

        codedeploy_app_name = resource_name(service.name + '-Application-Stack')
        cft.resources.add(Resource(codedeploy_app_name, 'AWS::CodeDeploy::Application'))
        cft.outputs.add(Output('CodeDeployApp', ref(codedeploy_app_name)))

        for stack in self.stacks:
            stack.to_cfn_template(service, self, template_bucket, codedeploy_app_name, cft, aws_manager, include_docker)

    def __read_extra_resources(self):
        if self.resources:
            try:
                with open(self.resources, 'r') as extra_resources_file:
                    rjson = json.load(extra_resources_file)
                    return [JSONableDict(update_dict=r[1], name=r[0]) for r in iter(rjson.items())]
            except IOError:
                raise NovaError("Environment resources '%s' not found." % self.resources)
        else:
            return None

    @staticmethod
    def load(values):
        return Environment(
            values.get("name"),
            values.get("aws_profile"),
            values.get("aws_region"),
            values.get("resources"),
            values.get("deploy_arn"),
            values.get("deployment_bucket"),
            values.get("deployment_application_id"),
            [Stack.load(s) for s in values.get("stacks")]
        )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
import os
import json
from botocore.waiter import WaiterModel, create_waiter_with_client
from nova.core import utils


class CloudformationWaiter(object):

    def __init__(self, client):
        waiter_json_filename = os.path.join(utils.__path__[0], 'cfn-waiters-2.json')
        with open(waiter_json_filename, 'r') as waiter_json_file:
            self.waiter_json_model = json.load(waiter_json_file)
        self.waiter_model = WaiterModel(self.waiter_json_model)
        self.waiter = create_waiter_with_client('StackAvailable', self.waiter_model, client.meta.client)



"""nova setup controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import shutil
from cement.core.controller import CementBaseController, expose

from nova.core import templates


class NovaSetupController(CementBaseController):
    class Meta:
        label = 'setup'
        stacked_on = 'base'
        stacked_type = 'nested'
        description = 'Setup NOVA scripts on a deployment host'
        usage = "nova setup"

    @expose(hide=True)
    def default(self):
        nova_install_dir = '/opt/nova'

        print("installing NOVA instance scripts, please wait...")
        instance_scripts_path = os.path.join(templates.__path__[0], 'instance-scripts')

        if not os.path.exists(nova_install_dir):
            os.makedirs(nova_install_dir)

        for f in os.listdir(instance_scripts_path):
            file_name = os.path.join(instance_scripts_path, f)
            shutil.copy(file_name, nova_install_dir)

        # This is a workaround so our tests don't see this as actual Nova Python source...
        os.rename(
            os.path.join(nova_install_dir, 'configure-cloudwatch-logs-agent.py.txt'),
            os.path.join(nova_install_dir, 'configure-cloudwatch-logs-agent.py')
        )

        print("NOVA instance scripts installed!")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import subprocess


class VersionManager(object):

    def __init__(self):
        self._current_version = None

    def current_version(self, forced_version=None):
        if self._current_version is None:
            if forced_version is None:
                git_version = subprocess.check_output(['git', 'describe', '--tags', '--dirty', '--always'])
                console_encoding = sys.stdout.encoding or 'utf-8'
                self._current_version = git_version.decode(console_encoding).strip('v').strip()
            else:
                self._current_version = forced_version
        return self._current_version



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from collections import OrderedDict
import pyaml
import json
from nova.core.cfn_pyplates.core import CloudFormationTemplate
from nova.core.exc import NovaError
from nova.core.spec.environment import Environment
from nova.core.spec.service_log_mapping import ServiceLogMapping


class Service(object):

    def __init__(self, name, team_name, port, healthcheck_url, logs, code_deploy_logs, environments):
        self.name = name
        self.team_name = team_name
        self.port = port
        self.healthcheck_url = healthcheck_url
        self.logs = logs
        self.code_deploy_logs = code_deploy_logs
        self.environments = environments

    def yaml(self):
        logs_list = None
        if self.logs is not None:
            logs_list = [lm.yaml() for lm in self.logs]

        data = OrderedDict([
            ('service_name', self.name),
            ('team_name', self.team_name),
            ('port', self.port),
            ('healthcheck_url', self.healthcheck_url),
            ('logs', logs_list),
            ('code_deploy_logs', self.code_deploy_logs),
            ('environments', [e.yaml() for e in self.environments])
        ])
        return pyaml.dump(OrderedDict((k, v) for k, v in data.items() if v is not None))

    def get_environment(self, environment_name):
        environment = next((item for item in self.environments if item.name == environment_name), None)
        if environment is not None:
            return environment
        else:
            raise NovaError("Environment '%s' was not found in the configuration" % environment_name)

    def set_code_deploy_app(self, environment_name, code_deploy_app_id):
        self.get_environment(environment_name).deployment_application_id = code_deploy_app_id

    def to_cfn_template(self, environment, template_bucket, aws_manager, cf_template_out=None, include_docker=True):
        description = "%s %s %s stack" % (self.team_name, self.name, environment.name)
        cfn = CloudFormationTemplate(description=description)
        environment.to_cfn_template(self, template_bucket, cfn, aws_manager, include_docker)
        json_cfn = cfn.json

        if cf_template_out is not None:
            with open(cf_template_out, 'w') as out_json:
                out_json.write(json.dumps(json.loads(json_cfn), indent=2))

        return json_cfn

    @staticmethod
    def load(values):
        logs_list = values.get("logs")
        if logs_list is not None:
            logs_list = [ServiceLogMapping.load(lm) for lm in logs_list]

        code_deploy_logs = True
        if values.get("code_deploy_logs") is not None:
            code_deploy_logs = values.get("code_deploy_logs")

        return Service(
            values.get("service_name"),
            values.get("team_name"),
            values.get("port"),
            values.get("healthcheck_url"),
            logs_list,
            code_deploy_logs,
            [Environment.load(e) for e in values.get("environments")]
        )



"""
nova main application entry point.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys
import traceback

from cement.core.foundation import CementApp
from cement.utils.misc import init_defaults
from cement.core.exc import FrameworkError, CaughtSignal
from cement.ext.ext_argparse import ArgParseArgumentHandler
from nova.core import exc

# this has to happen after you import sys, but before you import anything
# from Cement "source: https://github.com/datafolklabs/cement/issues/290"
if '--debug' in sys.argv:
    sys.argv.remove('--debug')
    TOGGLE_DEBUG = True
else:
    TOGGLE_DEBUG = False


# Application default.  Should update config/nova.conf to reflect any
# changes, or additions here.
defaults = init_defaults('nova')

# All internal/external plugin configurations are loaded from here
defaults['nova']['plugin_config_dir'] = '~/.nova/plugins.d'
# External plugins (generally, do not ship with application code)
defaults['nova']['plugin_dir'] = '~/.nova/plugins'
# External templates (generally, do not ship with application code)
defaults['nova']['template_dir'] = '~/.nova/templates'


class NovaArgHandler(ArgParseArgumentHandler):
    class Meta:
        label = 'nova_args_handler'

    def error(self, message):
        super(NovaArgHandler, self).error("unknown args")


class NovaApp(CementApp):
    class Meta:
        label = 'nova'
        config_defaults = defaults
        extensions = ['argcomplete']

        # All built-in application bootstrapping (always run)
        bootstrap = 'nova.cli.bootstrap'

        # Optional plugin bootstrapping (only run if plugin is enabled)
        plugin_bootstrap = 'nova.cli.plugins'

        # Internal templates (ship with application code)
        template_module = 'nova.core.templates'

        arg_handler = NovaArgHandler

        debug = TOGGLE_DEBUG


# Define the application object outside of main, as some libraries might wish
# to import it as a global (rather than passing it into another class/func)
app = NovaApp()


def main():
    with app:
        try:
            global sys
            # Default our exit status to 0 (non-error)
            code = 0

            # Dump all arguments into nova log
            app.log.debug(sys.argv)

            app.run()
        except exc.NovaError as e:
            # Catch our application errors and exit 1 (error)
            print('NovaError > %s' % e)
            code = 1

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('CaughtSignal > %s' % e)
            code = 0

        except FrameworkError as e:
            # Catch framework errors and exit 1 (error)
            print('FrameworkError > %s' % e)
            code = 1
        finally:
            # Print an exception (if it occurred) and --debug was passed
            if app.debug:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                if exc_traceback is not None:
                    traceback.print_exc()

        # Close the application
        app.close(code)


def get_test_app(**kw):
    app = NovaApp(**kw)
    return app


if __name__ == '__main__':
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class CustomScripts(object):

    def __init__(self,
                 app_stop_scripts=list(),
                 before_install_scripts=list(),
                 after_install_scripts=list(),
                 app_start_scripts=list(),
                 validate_scripts=list()):
        self.app_stop_scripts = app_stop_scripts
        self.before_install_scripts = before_install_scripts
        self.after_install_scripts = after_install_scripts
        self.app_start_scripts = app_start_scripts
        self.validate_scripts = validate_scripts

    @property
    def app_stop_scripts_for_app_spec(self):
        return [{'location': s} for s in self.app_stop_scripts]

    @property
    def before_install_scripts_for_app_spec(self):
        return [{'location': s} for s in self.before_install_scripts]

    @property
    def after_install_scripts_for_app_spec(self):
        return [{'location': s} for s in self.after_install_scripts]

    @property
    def app_start_scripts_for_app_spec(self):
        return [{'location': s} for s in self.app_start_scripts]

    @property
    def validate_scripts_for_app_spec(self):
        return [{'location': s} for s in self.validate_scripts]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from botocore.exceptions import ClientError, WaiterError
from termcolor import colored

from nova.core import check_latest_version
from nova.core.exc import NovaError
from nova.core.spec.nova_service_loader import NovaServiceLoader


class CreateStack:

    def __init__(self, aws_profile, environment_name, manager_provider,
                 cf_template_out=None, nova_descriptor_file=None, include_docker=True):
        check_latest_version()

        print("Creating cloudformation scripts...")

        self._environment_name = environment_name
        self._service_manager = NovaServiceLoader(environment_name, nova_descriptor_file)
        self._aws_manager = manager_provider.aws_manager(
            aws_profile or self._service_manager.environment.aws_profile,
            self._service_manager.environment.aws_region
        )

        self._s3_bucket = 'nova-deployment-templates-%s' % self._aws_manager.account_alias
        self.cloudformation_template = self._service_manager.service.to_cfn_template(
            self._service_manager.environment,
            self._s3_bucket,
            self._aws_manager,
            cf_template_out,
            include_docker
        )

    def create(self):
        try:
            self._aws_manager.create_and_upload_stack_template(self._s3_bucket, self._service_manager.service, self._service_manager.environment)

            self._aws_manager.create_stack(
                self._service_manager.service_name,
                self.cloudformation_template
            )

            stack = self._aws_manager.get_stack(self._service_manager.service_name)

            if stack.stack_status == "CREATE_COMPLETE":
                if stack.outputs is not None:
                    for output in stack.outputs:
                        if output.get('OutputKey') == 'CodeDeployApp':
                            code_deploy_app_id = output.get('OutputValue')
                            print(colored("Found code-deploy app: %s" % code_deploy_app_id, color='green'))
                            self._service_manager.service.set_code_deploy_app(self._environment_name, code_deploy_app_id)
                            print(colored("Please update nova.yml environment with 'deployment_application_id' manually.", color='yellow'))
                            break

                        for stack in self._service_manager.environment.stacks:
                            if output.get('OutputKey') == stack.name.title() + 'DeploymentGroup':
                                code_deploy_app_id = output.get('OutputValue')
                                print(colored("Found code-deploy deployment group for stack '%s': %s" % (stack.name, code_deploy_app_id), color='green'))
                                self._service_manager.service.set_code_deploy_app(self._environment_name, code_deploy_app_id)
                                print(colored("Please update nova.yml environment with 'deployment_application_id' manually.", color='yellow'))
                            else:
                                print(colored("Unable to find code-deploy deployment group for stack '%s' in the stack output" % stack.name, 'yellow'))
                                print(colored("Please update nova.yml stack with the 'deployment_group' manually.", color='yellow'))

                    else:
                        print(colored('Unable to find code-deploy application in the stack output', 'yellow'))
                        print(colored("Please update nova.yml environment with 'deployment_application_id'", color='yellow'))
                        print(colored("and your stacks with their respective 'deployment_group' manually.", color='yellow'))
                else:
                    print(colored("Please update nova.yml environment with 'deployment_application_id'", color='yellow'))
                    print(colored("and your stacks with their respective 'deployment_group' manually.", color='yellow'))

            else:
                raise NovaError("Stack creation was un-successful: %s" % stack.stack_status_reason)
        except ClientError as e:
            raise NovaError(str(e))
        except WaiterError as e:
            raise NovaError(str(e))



"""nova stash controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from cement.core.controller import CementBaseController, expose

from nova.core.managers.manager_provider import ManagerProvider
from nova.core.stash.decrypt import Decrypt
from nova.core.stash.encrypt import Encrypt
from nova.core.exc import NovaError

INCORRECT_GET_ARGS_USAGE = "Usage: nova stash get <environment> <key>"

INCORRECT_PUT_ARGS_USAGE = "Usage: nova stash put <environment> <key> <value>"


class NovaStashController(CementBaseController):
    class Meta:
        label = 'stash'
        description = 'NOVA service stash tools'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['-p', '--profile'], dict(help='AWS profile')),
            (['-r', '--region'], dict(help='AWS region')),
            (['-b', '--bucket'], dict(help='AWS bucket')),
            (['-k', '--key'], dict(help="AWS KMS Encryption Key Alias. (default is 'novastash')")),
            (['stash_args'], dict(action='store', nargs='*')),
        ]

    @expose(hide=True)
    def default(self):
        self.app.args.print_help()

    @expose(help='Decrypt a service variable')
    def get(self):
        if len(self.app.pargs.stash_args) == 1:
            profile = self.app.pargs.profile
            region = self.app.pargs.region
            bucket = self.app.pargs.bucket
            key = self.app.pargs.stash_args[0]
            manager_provider = ManagerProvider()
            Decrypt(
                stash_key=key,
                manager_provider=manager_provider,
                aws_profile=profile,
                aws_region=region,
                aws_bucket=bucket
            )
        else:
            raise NovaError(INCORRECT_GET_ARGS_USAGE)

    @expose(help='Encrypt a service variable')
    def put(self):
        if len(self.app.pargs.stash_args) == 2:
            profile = self.app.pargs.profile
            region = self.app.pargs.region
            bucket = self.app.pargs.bucket
            kms_key_alias = self.app.pargs.key or 'novastash'
            key = self.app.pargs.stash_args[0]
            value = self.app.pargs.stash_args[1]
            manager_provider = ManagerProvider()
            Encrypt(
                stash_key=key,
                value=value,
                manager_provider=manager_provider,
                aws_profile=profile,
                aws_region=region,
                aws_bucket=bucket,
                kms_key="alias/%s" % kms_key_alias
            )
        else:
            raise NovaError(INCORRECT_PUT_ARGS_USAGE)



"""nova stacks controller."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from cement.core.controller import CementBaseController, expose
from nova.core.exc import NovaError
from nova.core.managers.manager_provider import ManagerProvider
from nova.core.stack.create_stack import CreateStack
from nova.core.stack.update_stack import UpdateStack

INCORRECT_CREATE_ARGS_USAGE = "You must provide an environment to create"

INCORRECT_UPDATE_ARGS_USAGE = "You must provide an environment to update"


class NovaStacksController(CementBaseController):
    class Meta:
        label = 'stack'
        description = 'NOVA service stack tools'
        stacked_on = 'base'
        stacked_type = 'nested'
        arguments = [
            (['-f', '--file'], dict(help='Specify the nova service descriptor file to use.')),
            (['-p', '--profile'], dict(help='Override nova.yml AWS profile')),
            (['-o', '--output'], dict(help='Specify a file to output the template to.')),
            (['--with-docker-args'], dict(
                help='Populate legacy DockerDeployment* CF variables.',
                dest='include_docker',
                action='store_false',
                default=False
            )),
            (['environment'], dict(action='store', nargs='*'))
        ]
        usage = "nova stack [create|update] <environment>"

    @expose(hide=True)
    def default(self):
        self.app.args.print_help()

    @expose(help='Create NOVA service stack')
    def create(self):
        cf_template_out = self.app.pargs.output
        nova_descriptor_file = self.app.pargs.file
        include_docker = self.app.pargs.include_docker
        if self.app.pargs.environment:
            profile = self.app.pargs.profile
            CreateStack(
                aws_profile=profile,
                environment_name=self.app.pargs.environment[0],
                manager_provider=ManagerProvider(),
                cf_template_out=cf_template_out,
                nova_descriptor_file=nova_descriptor_file,
                include_docker=include_docker
            ).create()
        else:
            raise NovaError(INCORRECT_CREATE_ARGS_USAGE)

    @expose(help='Update NOVA service stack')
    def update(self):
        cf_template_out = self.app.pargs.output
        nova_descriptor_file = self.app.pargs.file
        include_docker = self.app.pargs.include_docker
        if self.app.pargs.environment:
            profile = self.app.pargs.profile
            UpdateStack(
                aws_profile=profile,
                environment_name=self.app.pargs.environment[0],
                manager_provider=ManagerProvider(),
                cf_template_out=cf_template_out,
                nova_descriptor_file=nova_descriptor_file,
                include_docker=include_docker
            ).update()
        else:
            raise NovaError(INCORRECT_UPDATE_ARGS_USAGE)
